/**
 * Hunter Adams (vha3@cornell.edu)
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 *  - GPIO 18 ---> 330 ohm resistor ---> VGA Green lo-bit |__ both wired to 150 ohm to ground 
 *  - GPIO 19 ---> 220 ohm resistor ---> VGA Green hi_bit |   and to VGA Green
 *  - GPIO 20 ---> 330 ohm resistor ---> VGA Blue
 *  - GPIO 21 ---> 330 ohm resistor ---> VGA Red
 *  - RP2040 GND ---> VGA GND
 *
 * RESOURCES USED for VGA
 *  - PIO state machines 0, 1, and 2 on PIO instance 0
 *  - DMA channels 0, 1, 2, and 3
 *  - 153.6 kBytes of RAM (for pixel color data)
 *
 * Protothreads v1.1.1
 * Serial console on GPIO 0 and 1 for debugging
 * Program will ru without a serial attach
 * 
 * Encoder:
 * ANO rotary encoder from Adafruit, Product ID: 5001
 * From the Adafruit docs, and oriented to that the connector is on the
 * Oriented so connnnector is at top. From right to left:
    ENCA - This is the rotary encoder A pin. ---> gpio 11
    ENCB - This is the rotary encoder B pin. ---> gpio 10
    COMA - Common pin A. connect to ground  
    SW1 - This pin is the center button.     ---> gpio 9
    SW2 - This pin is the down button.       ---> gpio 8
    SW3 - This pin is the right button.      ---> gpio 7
    SW4 - This pin is the up button.         ---> gpio 6
    SW5 - This pin is the left button.       ---> gpio 5
    COMB - Common pin B. connect to ground 
   
 * Joystick: 
 * HiLetgo Game Joystick Sensor Game Controller -- 
 * https://www.amazon.com/dp/B00P7QBGD2?psc=1&ref=ppx_yo2ov_dt_b_product_details
 * joystick 
 * pins:
 * GND ---> PICO GND
 * +5 ---> PICO +3
 * Rx ---> ADC channel 0 GPIO 26
 * Ry ---> ADC channel 0 GPIO 27
 * SW ---> GPIO 4 with pullup turned on
 * 
 * Noise level suggests only 8 bits of real resolution, not 12
 */

// ==========================================
// === VGA graphics library
// ==========================================
#include "vga16_graphics.h"
#include <stdio.h>
#include <stdlib.h>
// #include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/adc.h"
// // Our assembled programs:
// // Each gets the name <pio_filename.pio.h>
// #include "hsync.pio.h"
// #include "vsync.pio.h"
// #include "rgb.pio.h"

// ==========================================
// === hardware and protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "string.h"
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"

#define min(x,y) (((x)>(y))?(y):(x))
#define max(x,y) (((x)<(y))?(y):(x))

// joystick outputs
short adc_x, adc_y, button ;
short old_adc_x, old_adc_y ;
// encoder wheel and switches
short sw1, sw2, sw3, sw4, sw5, encA, encB ;
// encoder otuput (separate thread)
int encoder_count, encoder_delta ;

// ==================================================
// === graphics demo -- RUNNING on core 0
// ==================================================
static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);

    // main loop update count -- for tuning cursor speed
    static int update_count ;
    // encoder wheel count string
    static char vga_count_buffer[16] ;
    // encoder wheel count string
    static char vga_x_buffer[16] ;
    // encoder wheel count string
    static char vga_y_buffer[16] ;
    // the protothreads interval timer
    PT_INTERVAL_INIT() ;

    // ================== 
    // Banner setup
    // draw a big, clunky banner
    #define banner_bottom 50
    // Draw some filled rectangles at top of screen
    fillRect(64, 0, 176, banner_bottom, BLUE); // blue box
    fillRect(250, 0, 176, banner_bottom, YELLOW); // red box
    fillRect(435, 0, 176, banner_bottom, GREEN); // green box

    // Write some intro banner text
    setTextColor(WHITE) ;
    setCursor(65, 0) ;
    setTextSize(1) ;
    writeString("Raspberry Pi Pico") ;
    setCursor(65, 10) ;
    writeString("Encoder & joystick demo") ;
    setCursor(65, 20) ;
    writeString("Hunter Adams & Bruce Land") ;
    setCursor(65, 30) ;
    writeString("vha3@cornell.edu") ;
    setCursor(65, 40) ;
    writeString("brl4@cornell.edu") ;
    
    setTextColor(BLACK) ;
    setCursor(290, 6) ;
    setTextSize(2) ;
    writeString("ECE 4760") ;

    setCursor(445, 10) ;
    setTextSize(1) ;
    writeString("Protothreads rp2040 v1.1.1") ;
    setCursor(445, 20) ;
    writeString("VGA - 640x480 - 16 colors ") ;
    
    // ==================
    // Joystick ADC and button setup
    adc_init();
    // Make sure GPIO is high-impedance, no pullups etc
    // joystick x
    adc_gpio_init(26);
    // joystick y
    adc_gpio_init(27);
    // set up gpio 4 for joystick button
    #define joystick_button 4
    gpio_init(joystick_button) ;	
    gpio_set_dir(joystick_button, GPIO_IN) ;
    // pullup ON, pulldown OFF
    gpio_set_pulls (joystick_button, true, false) ;
    
    // ==================
    // Encoder inputs setup
    // gpio all set to input, pulled up
    for(int i=5; i<=11; i++){
        gpio_init(i) ;	
        gpio_set_dir(i, GPIO_IN) ;
        // pullup ON, pulldown OFF
        gpio_set_pulls (i, true, false) ;
    }

    // ==================
    // static graphics
    // draw a encoder icon 
    #define icon_center_x 150
    #define icon_center_y 250
    #define icon_radius 40
    drawCircle(icon_center_x, icon_center_y, icon_radius, YELLOW) ;
    setTextColor(WHITE) ;
    setCursor(icon_center_x-20, icon_center_y-15);
    setTextSize(1) ;
    writeString("encoder") ;

    // now the joystick icons
    #define joystick_base_x 350
    #define joystick_base_y 150
    #define joystick_width 129
    drawRect(joystick_base_x, joystick_base_y, joystick_width, 10, RED);
    drawRect(joystick_base_x-10, joystick_base_y+10, 10, joystick_width, RED);
    setTextColor(WHITE) ;
    setCursor(joystick_base_x+131, joystick_base_y-10);
    setTextSize(1) ;
    writeString("joystick x") ;
     setCursor(joystick_base_x-10, joystick_base_y+150) ;
    writeString("joystick y") ;

    // ==================
    // event loop
    while(true) {
        // count passes thru loop
        update_count++ ;
        
        // read joystick x, y potentiometers
        adc_select_input(0);
        old_adc_x = adc_x ;
        adc_x = adc_read();
        adc_select_input(1);
        old_adc_y = adc_y ;
        adc_y = adc_read();
        // read the joystick pushbutton
        button = !gpio_get(joystick_button);

        //  encoder wheel is read in a separate (blocking) thread
        //  and returns two values:
        //  encoder_count and encoder_delta
        // read encoder buttons 
        // center button
        sw1 = !gpio_get(9);
        // down
        sw2 = !gpio_get(8);
        // right
        sw3 = !gpio_get(7);
        // up
        sw4 = !gpio_get(6);
        // left
        sw5 = !gpio_get(5);
        
        // print all of the input variables
        //printf("%1d %1d %1d %1d %1d %5d %1d\n\r", sw1, sw2, sw3, sw4, sw5, encoder_count, encoder_delta);
        //printf("%1d %d %d\n\r\n\r", button, adc_x, adc_y);
        
        // animate encoder
        //erase old string
        setCursor(icon_center_x-20, icon_center_y-25) ;
        setTextSize(1) ;
        setTextColor(BLACK) ;
        writeString(vga_count_buffer) ;
        // write new one
        setTextColor(GREEN) ;
        setCursor(icon_center_x-20, icon_center_y-25) ;
        sprintf(vga_count_buffer, "%3d", encoder_count) ;
        writeString(vga_count_buffer) ;

        // 5 'LEDs' black/color for encoder switches
        fillCircle(icon_center_x, icon_center_y, 5, (sw1==true)*8) ;
        fillCircle(icon_center_x, icon_center_y-icon_radius, 5, (sw4==true)*5) ;
        fillCircle(icon_center_x, icon_center_y+icon_radius, 5, (sw2==true)*6) ;
        fillCircle(icon_center_x-icon_radius, icon_center_y, 5, (sw5==true)*7) ;
        fillCircle(icon_center_x+icon_radius, icon_center_y, 5, (sw3==true)*4) ;

        // draw a color square coresponding to the encoder modulo 15 when sw1 pushed
        if(sw1==true){
            fillRect(icon_center_x+icon_radius+20, icon_center_y-20, 40, 40, encoder_count%16) ;
        }
       
        // earase and redraw joystick x positions
        drawVLine(joystick_base_x+(old_adc_x>>5), joystick_base_y+1, 8, BLACK) ;
        drawVLine(joystick_base_x+(adc_x>>5), joystick_base_y+1, 8, RED) ;
        //erase old string
        setCursor(joystick_base_x+131, joystick_base_y) ;
        setTextSize(1) ;
        setTextColor(BLACK) ;
        writeString(vga_x_buffer) ;
        // write new one
        setTextColor(GREEN) ;
        setCursor(joystick_base_x+131, joystick_base_y) ;
        sprintf(vga_x_buffer, "%4d", adc_x) ;
        writeString(vga_x_buffer) ;

        // erase and redraw y
        drawHLine(joystick_base_x-10+1, joystick_base_y+10+(old_adc_y>>5), 8, BLACK) ;
        drawHLine(joystick_base_x-10+1, joystick_base_y+10+(adc_y>>5), 8, RED) ;
        setCursor(joystick_base_x-10, joystick_base_y+140) ;
        setTextSize(1) ;
        setTextColor(BLACK) ;
        writeString(vga_y_buffer) ;
        // write new one
        setTextColor(GREEN) ;
        setCursor(joystick_base_x-10, joystick_base_y+140) ;
        sprintf(vga_y_buffer, "%4d", adc_y) ;
        writeString(vga_y_buffer) ;

        // joystick button
        fillRect(joystick_base_x-10, joystick_base_y, 10, 10, (button==true)*8) ;
        // draw a small rect at location determined by joystick and with encoder wheel color
        // only when encoder middle buto pressed
        if(sw1==true){
            fillRect(joystick_base_x+(old_adc_x>>5)-5, joystick_base_y+(old_adc_y>>5)+5, 10, 10, BLACK) ;
            fillRect(joystick_base_x+(adc_x>>5)-5, joystick_base_y+(adc_y>>5)+5, 10, 10, encoder_count%16) ;
        }

        // run this about 16 times/sec
        PT_YIELD_usec(60000) ;
   }
   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
// just a heartbeat to make sure we did not crash
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     //gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ==================================================
// === user's serial input thread on core 1
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      
      //
      while(1) {
        // print prompt
        //sprintf(pt_serial_out_buffer, "just testing:");
        // spawn a thread to do the non-blocking write
        //serial_write ;

        // spawn a thread to do the non-blocking serial read
         //serial_read ;
         //printf("returned");
         PT_YIELD_usec(100000) ;
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ==================================================
// === encoder thread on core 1
// ==================================================
// read the encoder quadrature switches
static PT_THREAD (protothread_encoder(struct pt *pt))
{
    PT_BEGIN(pt);
    static int last_encA ;

    // wait for other core to start so that gpio is defined.
    PT_YIELD_usec(100000);
    // init the state to start with no count
    encA = !gpio_get(11);
    last_encA = encA;
    //
    while(1) {
        // most of the time this thread will be waiting
        // for a slow human event
        PT_YIELD_UNTIL(pt, !gpio_get(11) != last_encA);
        encA = !gpio_get(11);
        encB = !gpio_get(10);
        //if (encA != last_encA)
        do{     
            if (encB != encA) { 
                encoder_count ++;
                encoder_delta = +1;
            } else {
                encoder_count --;
                encoder_delta = -1;
            }
           //printf("%d\n\r", encoder_delta) ;
        } while(0);
        last_encA = encA ;
        // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // encoder thread


// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_serial) ;
  pt_add_thread(protothread_encoder) ; 
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");

  // Initialize the VGA screen
  initVGA() ;
     
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main